﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Manipulator {

  using HeuristicLab.Common;
  using HeuristicLab.Core;
  using HeuristicLab.Data;
  using HeuristicLab.Operators;
  using HeuristicLab.Optimization;
  using HeuristicLab.Parameters;
  using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
  using HeuristicLab.Problems.ProgramSynthesis.Base.Erc;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Configuration;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Encoding;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Generators.CodeGenerator;

  /// <summary>
  /// Uniformly distributed change of a single position of an integer vector.
  /// </summary>
  [Item("UniformMutation", " Uniformly distributed change of a single position of an plush vector.")]
  [StorableClass]
  public class UniformMutation : InstrumentedOperator, IPlushManipulator, IStochasticOperator {

    public UniformMutation() {
      Parameters.Add(new LookupParameter<IRandom>("Random", "The pseudo random number generator which should be used for stochastic manipulation operators."));
      Parameters.Add(new LookupParameter<PlushVector>("PlushVector", "The vector which should be manipulated."));
      Parameters.Add(new ValueLookupParameter<IReadOnlyExpressionsConfiguration>("Instructions", "The enabled instructions"));
      Parameters.Add(new ValueLookupParameter<IReadOnlyErcOptions>("ErcOptions", "ERC options"));
      Parameters.Add(new ValueLookupParameter<PercentValue>("InInstructionProbability", "The probability of IN Instructions"));

      Parameters.Add(new FixedValueParameter<PercentValue>("InstructionMutationProbability", "The probability an instruction gets mutated", new PercentValue(0.02)));
      Parameters.Add(new FixedValueParameter<PercentValue>("CloseMutationProbability", "The probability close gets mutated", new PercentValue(0.02)));
      Parameters.Add(new FixedValueParameter<PercentValue>("CloseIncrementRate", "When mutating individual, the increment rate specifies if close should be increased or decreased by 1.", new PercentValue(0.5)));
    }

    public UniformMutation(bool deserializing) : base(deserializing) {
    }

    public UniformMutation(UniformMutation origin, Cloner cloner) : base(origin, cloner) {
    }

    public IValueParameter<PercentValue> InstructionMutationProbabilityParameter
    {
      get { return (IValueParameter<PercentValue>)Parameters["InstructionMutationProbability"]; }
    }

    public double InstructionMutationProbability
    {
      get { return InstructionMutationProbabilityParameter.Value.Value; }
      set { InstructionMutationProbabilityParameter.Value.Value = value; }
    }

    public IValueParameter<PercentValue> CloseMutationProbabilityParameter
    {
      get { return (IValueParameter<PercentValue>)Parameters["CloseMutationProbability"]; }
    }

    public double CloseMutationProbability
    {
      get { return CloseMutationProbabilityParameter.Value.Value; }
      set { CloseMutationProbabilityParameter.Value.Value = value; }
    }

    public IValueParameter<PercentValue> CloseIncrementRateParameter
    {
      get { return (IValueParameter<PercentValue>)Parameters["CloseIncrementRate"]; }
    }

    public double CloseIncrementRate
    {
      get { return CloseIncrementRateParameter.Value.Value; }
      set { CloseIncrementRateParameter.Value.Value = value; }
    }

    public override bool CanChangeName
    {
      get { return false; }
    }
    public ILookupParameter<IRandom> RandomParameter
    {
      get { return (LookupParameter<IRandom>)Parameters["Random"]; }
    }
    public ILookupParameter<PlushVector> PlushVectorParameter
    {
      get { return (ILookupParameter<PlushVector>)Parameters["PlushVector"]; }
    }
    public IValueLookupParameter<IReadOnlyExpressionsConfiguration> InstructionsParameter
    {
      get { return (IValueLookupParameter<IReadOnlyExpressionsConfiguration>)Parameters["Instructions"]; }
    }
    public IValueLookupParameter<IReadOnlyErcOptions> ErcOptionsParameter
    {
      get { return (IValueLookupParameter<IReadOnlyErcOptions>)Parameters["ErcOptions"]; }
    }
    public IValueLookupParameter<PercentValue> InInstructionProbabilityParameter
    {
      get { return (IValueLookupParameter<PercentValue>)Parameters["InInstructionProbability"]; }
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new UniformMutation(this, cloner);
    }

    public sealed override IOperation InstrumentedApply() {
      Manipulate(
        RandomParameter.ActualValue,
        PlushVectorParameter.ActualValue,
        CloseIncrementRate,
        ErcOptionsParameter.ActualValue,
        InstructionsParameter.ActualValue,
        InInstructionProbabilityParameter.ActualValue.Value,
        CloseMutationProbability,
        InstructionMutationProbability);
      return base.InstrumentedApply();
    }

    private static void Manipulate(
      IRandom random,
      PlushVector plushVector,
      double closeIncrementRate,
      IReadOnlyErcOptions ercOptions,
      IReadOnlyExpressionsConfiguration instructions,
      double inInstructionProbability,
      double closeMutationProbability,
      double instructionMutationProbability
      ) {

      for (var i = 0; i < plushVector.Entries.Count; i++) {
        var entry = plushVector[i];
        var x = random.NextDouble();

        if (x < instructionMutationProbability) {
          var instruction = CodeGeneratorUtils.GetRandomExpression(
            random,
            ercOptions,
            instructions,
            inInstructionProbability);
          entry.Instruction = instruction;
        }

        x = random.NextDouble();

        if (x < closeMutationProbability) {
          x = random.NextDouble();
          entry.Close += x < closeIncrementRate ? 1 : -1;

          if (entry.Close < 0)
            entry.Close = 0;
        }
      }

      // specifies that the program cached within entry must be updated before reuse
      plushVector.UpdatePushProgram();
    }
  }
}
